#ifndef _XRDDPMXMI_H_
#define _XRDDPMXMI_H_
/******************************************************************************/
/*                                                                            */
/*                          X r d D P M X m i . h h                           */
/*                                                                            */
/* DPM handling and various by David Smith, CERN                              */
/* Other parts:                                                               */
/*                                                                            */
/* (c) 2007 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/******************************************************************************/

#include "XrdCms/XrdCmsReq.hh"
#include "XrdCms/XrdCmsXmi.hh"
#include "XrdSys/XrdSysPthread.hh"
#include "XrdAcc/XrdAccAuthorize.hh"

#include "dpns_api.h"
#include "serrno.h"

#include <memory>

class XrdSysError;
class XrdInet;
class XrdScheduler;
class XrdOucTrace;
class XrdDPMQueue;

namespace DPMXmi {
class XmiErr;
}

class XrdDPMXmi : XrdCmsXmi
{
public:

       int  Chmod (       XrdCmsReq      *Request,
                          mode_t          mode,
                    const char           *path,
                    const char           *opaque);

       int  Mkdir (       XrdCmsReq      *Request,
                          mode_t          mode,
                    const char           *path,
                    const char           *opaque);

       int  Mkpath(       XrdCmsReq      *Request,
                          mode_t          mode,
                    const char           *path,
                    const char           *opaque);

       int  Prep  ( const char           *ReqID,
                          int             Opts,
                    const char           *Path,
                    const char           *Opaque)
                  {return 1;}

       int  Rename(       XrdCmsReq      *Request,
                    const char           *oldpath,
                    const char           *oldopaque,  
                    const char           *newpath,
                    const char           *newopaque);

       int  Remdir(       XrdCmsReq      *Request,
                    const char           *path,
                    const char           *opaque);

       int  Remove(       XrdCmsReq      *Request,
                    const char           *path,
                    const char           *opaque);

       int  Select(       XrdCmsReq      *Request,
                          int             opts,
                    const char           *path,
                    const char           *opaque);

       int  Stat  (       XrdCmsReq      *Request,
                    const char           *path,
                    const char           *opaque);

       // processing methods called from a XmiDPMReq object
       //
       int      Chmod_process(XrdDPMReq *rp, const bool &destroy);
       int      Mkdir_process(XrdDPMReq *rp, const bool &destroy);
       int      Mkpath_process(XrdDPMReq *rp, const bool &destroy);
       int      Rename_process(XrdDPMReq *rp, const bool &destroy);
       int      Remdir_process(XrdDPMReq *rp, const bool &destroy);
       int      Remove_process(XrdDPMReq *rp, const bool &destroy);
       int      Select_process(XrdDPMReq *rp, const bool &destroy);
       int      Stat_process(XrdDPMReq *rp, const bool &destroy);

       // for getting request specific options from the surl
       //
       int      getExtraSurlOptions(const char *p, const char *key, size_t maxlen,
                                    const char **outp, size_t *outlen) const;
       int      getExtraSurlOptions(const char *Path, XrdDPMReq *rp) const;

       // default lfn2rfn for use by the XmiDPMQueue object
       static int      defaultLfn2rfn(const char *Path, char *new_path, size_t s);

       void     InitXeq();

       void     MSSPoll();

                XrdDPMXmi(XrdCmsXmiEnv *);

                ~XrdDPMXmi();

       // This XeqMode is the same as the default virtual method in the base class,
       // but we state it explictly in the XrdDPMXmi class
       // 
       void     XeqMode ( unsigned int &isNormal,
                          unsigned int &isDirect)
                    {isNormal = XMI_LOAD | XMI_SPACE; isDirect = 0;}

private:

       // structure with various DPM options: filled from configuration command line args and file
       //

const static int MaxFqans = 32;

struct dpmoptions_s {
          char           dpns_host[CA_MAXHOSTNAMELEN+1];
          unsigned short dpns_port;
          unsigned short xrd_server_port;
          bool           mkpath_bool;
          time_t         reqput_lifetime;
          char           reqput_ftype;
          char           reqput_stoken[CA_MAXDPMTOKENLEN+1];
          u_signed64     reqput_reqsize;
          time_t         reqget_lifetime;
          char           reqget_ftype;
          char           reqget_stoken[CA_MAXDPMTOKENLEN+1];
          char           principal[CA_MAXCSECNAMELEN+1];
          char           voname[CA_MAXCSECNAMELEN+1];
          int            nbfqans;
          char           fqans[MaxFqans][CA_MAXCSECNAMELEN+1];
          char           authlib[1024];
          char           authparm[1024];
          int            authopt;
          XrdAccAuthorize* authorization;
       }             Opts;
       //
       // structure for private data for processing chmod, mkdir, mkpath
       //
struct pData1 {
          mode_t         mode;
       };
       //
       // structure for private data for rename
       //
struct pData2 {
          char           opath[CA_MAXPATHLEN+1];
          char           npath[CA_MAXPATHLEN+1];
       };
       //                                                                               
       // structure for private data for select (i.e. open)
       //
struct pData3 {
          int            opts;
          bool           tried_mkpath;
          bool           get,first;
          char           r_token[CA_MAXDPMTOKENLEN+1];
          char           fstatbuf[256];
       };

       int      Select_process_put_first(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error);
       int      Select_process_put(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error);
       int      Select_process_get_first(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error);
       int      Select_process_get(XrdDPMReq *rp, std::auto_ptr<pData3> &pd, DPMXmi::XmiErr &error);

       void     Init(int When=0, int force=0);
       int      ageSession();
       int      ageSession(const time_t &t);
       int      setupAndUseSession(XrdDPMReq *rp, DPMXmi::XmiErr &err);
       int      setupAndUseSession();
       int      closeSession();
       int      mkp(const char *path, const mode_t &mode) const;
       int      SetDefaultsAndReadConfig(XrdCmsXmiEnv *Env);
const  char    *rcToStr(const int &rc) const;
       int      sendError(XrdCmsReq *reqP, const int &rc, const char *opn, const char *path) const;
       int      sendWait(XrdCmsReq *reqP, const int &rc, const char *opn) const;
       void     sendRedirect(XrdCmsReq *Request, const char *turl, const char *token, const bool &put) const;
       int      notSupported(XrdCmsReq *rp, const char *opn, const char *path) const;

       // Various pointers
       //
static XrdOucName2Name *N2N;            // -> lfn to xxx mapper (may be null)
static XrdSysError  *eDest;             // -> Error message handler
static XrdInet      *iNet;              // -> Network object
static XrdScheduler *Sched;             // -> Thread scheduler
static XrdOucTrace  *Trace;             // -> Trace handler

       // some constants
       //
static const int     retryTime        =  90; // How long a client waits between retries
static const int     MSSPollTime      =  30; // Poll waiting requests at least every 30 seconds
static const int     maxsessage       =  20; // DPNS session becomes liable to be closed after 20 seconds inactivity

       // Initialization control variable protected by initMutex
       //
       XrdSysMutex   initMutex;
       char          initDone;
       char          initActive;

       // Error buffer for dpns/dpm API
       char          m_apibuf[1024];

       // endpoint, dpns session status and request queue
       //
       char          m_ServerAndPort[CA_MAXHOSTNAMELEN+7];
       bool          m_dpnsSessEst;
       time_t        m_dpnsSessLastChecked;
       XrdDPMQueue  *queue;
};

/******************************************************************************/
/*                            N a m e M a p p e r                             */
/******************************************************************************/
class XrdDPMXmiN2N : public XrdOucName2Name
{
public:
       	int lfn2pfn(const char *lfn, char *buff, int blen);
        int lfn2rfn(const char *lfn, char *buff, int blen);
        int pfn2lfn(const char *pfn, char *buff, int blen);

            XrdDPMXmiN2N(XrdSysError *erp, const char *lpfx, const char *rpfx);
       int set_mad(const char *mad);

private:

XrdSysError *eDest;
};
#endif
